unit UnitScrollingLabel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls;

Const  Shareware=False;
              Id='c-FD 97';
             Ver='1.01';
type
  TScreenMode=(SmNormal,SmDot);
  TScrollingLabel = class(TCustomPanel)
  private
    { Dclarations prives }
    Bmp1:TBitmap;   // Bitmap contenant l'image d'origine
    FOffset:Integer;   // Position d'origine de l'affichage dans Bitmap d'origine
    Bmp2:TBitmap;   // Bitmap contenant l'image a afficher en tenant compte de l'offset
    Bmp3:TBitmap;
    Horloge:TTimer; // Horloge pour defilement Auto
    TailleAff,TailleSource:TRect;
    FBackGroundColor:TColor;
    FTextColor:TColor;
    FOffColor:TColor;
    FCaption:String;
    FTextFontName:TFontName;
    FTextSize:Integer;
    FTextStyle:TFontStyles;
    FScroll:Boolean; // A True pour defilement auto
    FMaxOffset:Word; // Offset maxi
    FScrollInterval:Word; // Valeur pour l'horloge
    FScrollStep:Integer; // Pas du scrolling auto
    FScreenMode:TScreenMode; // Affichage normal ou en mode point
    FDotSize:Byte; // Taille des points pour affichage en mode point
    FBlankSize:Byte;
    FGoAndBack:Boolean;
    FOnZeroPos:TNotifyEvent;
    FVer,Dummy:String;
    Procedure SetCaption(ACaption:String);
    Procedure SetBackGroundColor(ABackGroundColor:TColor);
    Procedure SetTextColor(ATextColor:TColor);
    Procedure SetOffColor(AOffColor:TColor);
    Procedure SetTextFontName(ATextFontName:TFontName);
    Procedure SetTextSize(ATextSize:Integer);
    Procedure SetTextStyle(ATextStyle:TFontStyles);
    Procedure SetOffset(AOffset:Integer);
    Procedure SetScroll(AScroll:Boolean);
    Procedure SetScrollInterval(AScrollInterval:Word);
    Procedure SetScrollStep(AScrollStep:Integer);
    Procedure SetScreenMode(AScreenMode:TScreenMode);
    Procedure SetDotSize(ADotSize:Byte);
    Procedure SetBlankSize(ABlankSize:Byte);
    Procedure ClearControl;
    Procedure Paint; Override;
    Procedure InitBitmap(St:String);
    Procedure ActualiseBmp2;
    Procedure Scrolling(Sender:TObject); // Fait scroller l'affichage
    Procedure Resize; Override; // Surcharge le methode
  protected
    { Dclarations protges }
    Procedure Free;
  public
    { Dclarations publiques }
    Property MaxOffset:Word Read FMaxOffset;
    Constructor Create(AOwner:TComponent); Override;
  published
    { Dclarations publies }
    Property Caption:String Read FCaption Write SetCaption;
    Property BackGroundColor:TColor Read FBackGroundColor Write SetBackGroundColor;
    Property TextColor:TColor Read FTextColor Write SetTextColor;
    Property LedOffColor:TColor Read FOffColor Write SetOffColor;
    Property TextFontName:TFontName Read FTextFontName Write SetTextFontName;
    Property TextSize:Integer Read FTextSize Write SetTextSize;
    Property TextStyle:TFontStyles Read FTextStyle Write SetTextStyle;
    Property Offset:Integer Read FOffset Write SetOffset;
    Property Scroll:Boolean Read FScroll Write SetScroll;
    Property Scroll_Interval:Word Read FScrollInterval Write SetScrollInterval;
    Property Scroll_Step:Integer Read FScrollStep Write SetScrollStep;
    Property Screen_Mode:TScreenMode Read FScreenMode Write SetScreenMode;
    Property Dot_Size:Byte Read FDotSize Write SetDotSize;
    Property Blank_Size:Byte Read FBlankSize Write SetBlankSize;
    Property OnZeroPos:TNotifyEvent Read FOnZeroPos Write FOnZeroPos;
    Property GoAndBack:Boolean Read FGoAndBack Write FGoAndBack;
    Property Version:String Read FVer Write Dummy;
    Property ShowHint;
  end;

procedure Register;

implementation

{----------------------------------------------------------}
{ Constructeur du composant }
Constructor TScrollingLabel.Create(AOwner:TComponent);

begin
  Inherited Create(AOwner);
  Bmp1:=TBitmap.Create;
  Bmp2:=TBitmap.Create;
  Bmp3:=TBitmap.Create;
  FCaption:=Id;
  FVer:=Ver+'  '+Id;
  FBackGroundColor:=ClBtnFace;
  FTextFontName:='Arial';
  FTextColor:=ClBtnText;
  FOffColor:=ClSilver;
  FTextSize:=10;
  Width:=55;Height:=25;
  FOffset:=0;
  FScroll:=False;
  FScrollInterval:=250;
  FScrollStep:=2;
  FScreenMode:=SmNormal;
  FDotSize:=2;
  FBlankSize:=1;
  FGoAndBack:=False;
  Horloge:=TTimer.Create(Self);
  With Horloge do begin
                    Enabled:=FScroll;
                    Interval:=FScrollInterval;
                    OnTimer:=Scrolling;
                  end;
 InitBitmap(FCaption);
end;
{----------------------------------------------------------}
{ Libere le composant et ses ressources }
Procedure TScrollingLabel.Free;

begin
  Horloge.Free;
  Bmp1.Free;
  Bmp2.Free;
  Bmp3.Free;  // Libere les ressources
  Inherited Free;  // Appele la methode heritee
end;
{----------------------------------------------------------}
{ Efface la zone du controle }
Procedure TScrollingLabel.ClearControl;

begin
  Canvas.Brush.Color:=FBackGroundColor;
  Canvas.FillRect(Rect(0,0,Width,Height)); // Efface l'ancien controle
end;
{----------------------------------------------------------}
{ Realise un bitmap aux dimensions de la chaine passee en parametre }
Procedure TScrollingLabel.InitBitmap(St:String);
Var       Ht,Lg:Word;
         BmpTmp:TBitmap;
            X,Y:Word;

begin
  With Bmp1.Canvas do
  begin
    Brush.Color:=FBackGroundColor;
    FillRect(Rect(0,0,Width,Height)); // Efface l'ancien controle
    Font.Color:=FTextColor;
    Font.Name:=FTextFontName;
    Font.Style:=FTextStyle;
    Font.Size:=FTextSize;
    Bmp1.Height:=TextHeight(St);
    if FScreenMode=SmDot then While TextWidth(St)<Width/(FDotSize) do st:=st+' '
                         else While TextWidth(St)<Width do st:=st+' ';
    Bmp1.Width:=TextWidth(St);
    TextOut(0,0,St);
  end; // With Bmp1.Canvas
  If FScreenMode=SmDot then // Si on est en mode point
  begin
    BmpTmp:=TBitmap.Create; // Creation d'un bitmap temporaire
    BmpTmp.Height:=(Bmp1.Height*FDotSize)+((Bmp1.Height-1)*1);
    BmpTmp.Width:=(Bmp1.Width*FDotSize)+((Bmp1.Width-1)*1);
    While BmpTmp.Width div (FDotSize+FBlankSize)<>BmpTmp.Width/(FDotSize+FBlankSize) do BmpTmp.Width:=BmpTmp.Width-1;
    BmpTmp.Canvas.Brush.Color:=FBackGroundColor;
    BmpTmp.Canvas.FillRect(Rect(0,0,BmpTmp.Width,BmpTmp.Height));
    For X:=0 To Bmp1.Width-1 do
        For Y:=0 to Bmp1.Height-1 do
            begin
              If Bmp1.Canvas.Pixels[X,Y]=FTextColor then
              begin // Si le point est a afficher
                BmpTmp.Canvas.pen.color:=FTextColor;  // On l'affiche couleur du texte
                BmpTmp.Canvas.Brush.Color:=FTextColor;
              end else
              begin
                BmpTmp.Canvas.pen.color:=FOffColor; // Sinon on l'affiche couleur du fond
                BmpTmp.Canvas.Brush.Color:=FOffColor;
              end; // If
              // Dessine le point en forme de cercle
              BmpTmp.Canvas.Ellipse(X*(FDotSize+FBlankSize)-(FDotSize div 2),   // X1
                                    Y*(FDotSize+FBlankSize)-(FDotSize div 2),   // Y1
                                    X*(FDotSize+FBlankSize)+(FDotSize div 2),   // X2
                                    Y*(FDotSize+FBlankSize)+(FDotSize div 2))   // Y2
            end; // For
    Bmp1.Height:=BmpTmp.Height;Bmp1.Width:=BmpTmp.Width;
    Bmp1:=BmpTmp; // on passe BmpTmp dans Bmp1
  end; // If FScreenMode=SmDot
  FMaxOffset:=Bmp1.Width;
  Bmp2.Height:=Bmp1.Height;Bmp2.Width:=Bmp1.Width; // Le 2eme Bitmap a les mme dimension que le 1er
  With TailleAff do begin
                      Top:=(Height div 2)-(Bmp2.Height div 2); // On Centre l'affichage dans le
                      Left:=0;                                 // composant
                      Right:=Width;
                      Bottom:=Top+Bmp2.Height;
                    end;
  With TailleSource do begin
                         Top:=0;
                         Left:=0;
                         Right:=Width;
                         Bottom:=Bmp2.Height;
                       end;
end;
{----------------------------------------------------------}
{ Actualise Bmp2 ( Bmp2 contient l'image a afficher ) }
Procedure TScrollingLabel.ActualiseBmp2;

begin
  If Offset=0 then Bmp2.Canvas.CopyRect(Rect(0,0,Bmp2.Width,Bmp2.Height),Bmp1.Canvas,Rect(0,0,Bmp2.Width,Bmp2.Height)) // On fait juste une copie
              else begin // Sinon on copie en faisant une rotation
                     Bmp2.Canvas.CopyRect(Rect(0,0,Bmp2.Width-FOffset,Bmp2.Height),Bmp1.Canvas,Rect(FOffset,0,Bmp2.Width,Bmp2.Height));
                     Bmp2.Canvas.CopyRect(Rect(Bmp2.Width-FOffset,0,Bmp2.Width,Bmp2.Height),Bmp1.Canvas,Rect(0,0,FOffset,Bmp2.Height))
                   end;
end;
{----------------------------------------------------------}
{ Nouvelle methode Paint }
Procedure TScrollingLabel.Paint;

begin
  ActualiseBmp2; // Effectue une rotation si offset<>0
  Canvas.CopyRect(TailleAff,Bmp2.Canvas,TailleSource);
end;
{----------------------------------------------------------}
{ Fait scroller l'affichage }
Procedure TScrollingLabel.Scrolling(Sender:TObject);
var       Event:Boolean;

begin
  Inc(FOffset,FScrollStep);Event:=False;
  If (FScrollStep>0) and (FOffset>=MaxOffset) then begin
                                                     FOffset:=0;
                                                     Event:=True
                                                   end;
  If (FScrollStep<0) and (FOffset<=0) then begin
                                             FOffset:=MaxOffset;
                                             Event:=True
                                           end;
  if Event then
  begin         // Si passage a zero
    if FGoAndBack then begin  // On change de sens de defilement
                         Scroll_Step:=-Scroll_Step;
                         if Scroll_Step>0 then Offset:=1;
                         if Scroll_Step<0 then Offset:=MaxOffset-1;
                       end;
    If Assigned(FOnZeroPos) then FOnZeroPos(Self); // On appele l'evenement
  end;
  Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete Caption }
Procedure TScrollingLabel.SetCaption(ACaption:String);

begin
  If ACaption=FCaption then Exit;
  FCaption:=ACaption;
  If Shareware then FCaption:=Id+Fcaption;
  InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete BackGroundColor }
Procedure TScrollingLabel.SetBackGroundColor(ABackGroundColor:TColor);

begin
  If ABackGroundColor=FBackGroundColor then Exit;
  FBackGroundColor:=ABackGroundColor;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete Font }
Procedure TScrollingLabel.SetTextStyle(ATextStyle:TFontStyles);

begin
  If ATextStyle=FTextStyle then Exit;
  FTextStyle:=ATextStyle;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete TextColor }
Procedure TScrollingLabel.SetTextColor(ATextColor:TColor);

begin
  If ATextColor=FTextColor then Exit;
  FTextColor:=ATextColor;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete OffColor }
Procedure TScrollingLabel.SetOffColor(AOffColor:TColor);

begin
  If AOffColor=FOffColor then Exit;
  FOffColor:=AOffColor;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete Font }
Procedure TScrollingLabel.SetTextFontName(ATextFontName:TFontName);

begin
  If ATextFontName=FTextFontName then Exit;
  FTextFontName:=ATextFontName;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete TextFontSize }
Procedure TScrollingLabel.SetTextSize(ATextSize:Integer);

begin
  If ATextSize=FTextSize then Exit;
  FTextSize:=ATextSize;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete Offset }
Procedure TScrollingLabel.SetOffset(AOffset:Integer);

begin
  If AOffset=FOffset then Exit;
  If AOffset>MaxOffset then AOffset:=0;
  If AOffset<0 then AOffset:=MaxOffset;
  FOffset:=AOffset;
  Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete Scroll }
Procedure TScrollingLabel.SetScroll(AScroll:Boolean);

begin
  If AScroll=FScroll then Exit;
  FScroll:=AScroll;
  Horloge.Enabled:=FScroll;
end;
{----------------------------------------------------------}
{ Traite la propriete ScrollInterval }
Procedure TScrollingLabel.SetScrollInterval(AScrollInterval:Word);

begin
  If AScrollInterval=FScrollInterval then Exit;
  If (AScrollInterval<>0) and (AScrollInterval<10) then AScrollInterval:=10;
  FScrollInterval:=AScrollInterval;
  Horloge.Interval:=FScrollInterval;
end;
{----------------------------------------------------------}
{ Traite la propriete ScrollStep }
Procedure TScrollingLabel.SetScrollStep(AScrollStep:Integer);

begin
  If AScrollStep=FScrollStep then Exit;
  If AScrollStep=0 then AScrollStep:=1;
  If AScrollStep>MaxOffset then AScrollStep:=MaxOffset;
  If -AScrollStep<-MaxOffset then AScrollStep:=-MaxOffset;
  FScrollStep:=AScrollStep;
end;
{----------------------------------------------------------}
{ Traite la propriete ScreenMode }
Procedure TScrollingLabel.SetScreenMode(AScreenMode:TScreenMode);

begin
  If AScreenMode=FScreenMode then Exit;
  FScreenMode:=AScreenMode;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete DotSize }
Procedure TScrollingLabel.SetDotSize(ADotSize:Byte);

begin
  If ADotSize=FDotSize then Exit;
  If ADotSize<2 then ADotSize:=2;
  If ADotSize>20 then ADotSize:=20;
  FDotSize:=ADotSize;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Traite la propriete BlankSize }
Procedure TScrollingLabel.SetBlankSize(ABlankSize:Byte);

begin
  If ABlankSize=FBlankSize then Exit;
  If ABlankSize=0 then ABlankSize:=1;
  If ABlankSize>20 then ABlankSize:=20;
  FBlankSize:=ABlankSize;
  ClearControl;InitBitmap(FCaption);Paint;
end;
{----------------------------------------------------------}
{ Surcharge la methode Resize }
Procedure TScrollingLabel.Resize;

begin
  InitBitmap(FCaption);Paint; // Appele la nouvelle methode
  Inherited;
end;
{----------------------------------------------------------}
{ Enregistre le composant }
procedure Register;
begin
  RegisterComponents('FD Components', [TScrollingLabel]);
end;

end.
