{
Copyright (C) 2002-2009  Massimo Melina (www.rejetto.com)

This file is part of Http File Server (HFS).

    HFS is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    HFS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with HFS; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit traylib;

interface

uses
  forms, ShellAPI, Messages, windows, graphics, sysutils, classes;

const
  WM_TRAY = WM_USER+1;
type
  TtrayEvent = (TE_CLICK, TE_2CLICK, TE_RCLICK);

  TNotifyIconDataEx = record
    cbSize: DWORD;
    wnd: HWND;
    uID: UINT;
    uFlags: UINT;
    uCallbackMessage: UINT;
    hIcon: HICON;
    szTip: array [0..127] of AnsiChar;
    dwState: DWORD;
    dwStateMask: DWORD;
    szInfo: array[0..255] of AnsiChar;
    uVersion: UINT;
    szInfoTitle: array[0..63] of AnsiChar;
    dwInfoFlags: DWORD;
    end;

  TtrayIcon=class
    private
      icondata: TNotifyIconDataEx;
      shown: boolean;
      procedure wndProc(var Message: TMessage);
      procedure notify(ev:TtrayEvent);
    public
      data: pointer;  // user data
      onEvent: procedure(sender:Tobject; ev:TtrayEvent) of object;
      constructor create(form:Tform);
      destructor Destroy; override;
      procedure minimize;
      procedure update;
      procedure hide;
      procedure show;
      procedure setIcon(icon:Ticon);
      procedure setTip(s:string);
      procedure setIconFile(fn:string);
      procedure updateHandle(handle:HWND);
    end; // TtrayIcon

implementation

var
  maxTipLength: integer;

constructor TtrayIcon.create(form:Tform);
begin
with icondata do
  begin
  uCallbackMessage := WM_TRAY;
  cbSize := sizeof(icondata);
  Wnd := classes.AllocateHWnd(wndproc);
  uID := 100;
  uFlags := NIF_MESSAGE + NIF_ICON + NIF_TIP;
  end;
setIcon(application.icon);
setTip(application.title);
end; // create

destructor Ttrayicon.destroy;
begin
classes.DeallocateHWnd(icondata.wnd);
hide;
end;

procedure TtrayIcon.updateHandle(handle:HWND);
begin
if not shown then
  begin
  icondata.wnd:=handle;
  exit;
  end;
hide;
icondata.wnd:=handle;
Shell_NotifyIcon(NIM_ADD, @icondata)
end;

procedure TtrayIcon.update;
begin
if shown then
  if not Shell_NotifyIcon(NIM_MODIFY, @icondata) then
    Shell_NotifyIcon(NIM_ADD, @icondata);
end; { update }

procedure TtrayIcon.setIcon(icon:Ticon);
begin
if icon=NIL then exit;
icondata.hIcon:=icon.handle;
update;
end; { setIcon }

procedure TtrayIcon.setIconFile(fn:string);
var
  ico:Ticon;
begin
ico:=Ticon.create;
ico.loadFromFile(fn);
setIcon(ico);
end; // setIconFile

procedure TtrayIcon.setTip(s:string);
begin
s:=stringReplace(s,'&','&&',[rfReplaceAll]);
if length(s) > maxTipLength then setlength(s,maxTipLength);
if icondata.szTip = s then exit;
strPCopy(icondata.szTip, s);
update();
end; // setTip

procedure TtrayIcon.minimize;
begin
show;
Application.ShowMainForm := False;
// Toolwindows dont have a TaskIcon. (Remove if TaskIcon is to be show when form is visible)
SetWindowLong(Application.Handle, GWL_EXSTYLE, WS_EX_TOOLWINDOW);
end; // minimizeToTray

procedure TtrayIcon.show;
begin
if shown then exit;
shown:=true;
Shell_NotifyIcon(NIM_ADD, @icondata)
end; // show

procedure TtrayIcon.hide;
begin
if not shown then exit;
shown:=FALSE;
Shell_NotifyIcon(NIM_DELETE, @icondata)
end; // hide

procedure TtrayIcon.wndproc(var Message: TMessage);
begin
case message.Msg of
  WM_TRAY:
    case message.lParam of
      WM_RBUTTONUP: notify(TE_RCLICK);
      WM_LBUTTONUP: notify(TE_CLICK);
      WM_LBUTTONDBLCLK: notify(TE_2CLICK);
      end;
  end;
message.result:=1;
end;

procedure TtrayIcon.notify(ev:TtrayEvent);
begin if assigned(onEvent) then onEvent(self, ev) end;

INITIALIZATION
if byte(getVersion()) < 5 then maxTipLength:=63
else maxTipLength:=127;

end.
